#version 140

#ifdef GL_ES
precision highp float;
#endif

out vec4 fragColor;

// note: clipping disabled for volumetric fog because it interferes with mirrors
//#include "../include/clip_plane.frag"

in vec3 world_dir;

// depth
uniform sampler2D tex5;

// xy = positive 1/screen dims
// zw = projection matrix components
uniform vec4 ofs;

// these two specify frustum corners
uniform vec3 corner;
uniform vec3 cornerDelta;

uniform vec4 light_color;

uniform vec3 fog_pos_rel;
uniform vec3 fog_extent;

uniform vec3 fog_box_min_rel;
uniform vec3 fog_box_max_rel;

uniform float fog_density;
// (1, 0) = exp, (0, 1) = exp2
uniform vec2 fog_falloff;

float LinearDepth(float d)
{
	// now with pre-baked denormalization and sign flip
	return ofs.w / (ofs.z + d);
}

#ifdef FOG_SPHERE

vec2 SolveQuadraticEquation(float a, float b, float c)
{
	// in our case, a is always > 0
	float a2 = 2.0*a;

	float D = b*b - 4.0*a*c;

	// this should be faster than if; it will give us two same t's => density of 0, which is what we want
	D = max(D, 0.0);

	float sqD = sqrt(D);

	float invA2 = 1.0/a2;

	float x1 = (-b + sqD) * invA2;
	float x2 = (-b - sqD) * invA2;

	// min/max rather than conditional
	return vec2(min(x1, x2), max(x1, x2));
}

vec2 IntersectUnitSphereAtOrigin(vec3 o, vec3 d)
{
	float a = dot(d, d);
	float b = 2.0 * dot(o, d);
	// unit sphere so radius^2 is 1
	float qc = dot(o, o) - 1.0;

	return SolveQuadraticEquation(a, b, qc);
}

float GetSpherePenetration(vec3 ray_dir, vec3 org, vec3 rad, float maxt)
{
	// we're raycasting ray at origin
	// transform to unit sphere
	vec3 inv_rad = vec3(1.0) / rad;
	ray_dir *= inv_rad;
	org *= inv_rad;

	// now we need to intersect unit sphere at org
	vec2 t = IntersectUnitSphereAtOrigin(-org, ray_dir);

	t[0] = max(t[0], 0.0);
	t[1] = min(t[1], maxt);

	return max(0.0, t[1] - t[0]);
}

#else

// reference: https://tavianator.com/fast-branchless-raybounding-box-intersections-part-2-nans/
// note: ray starts at origin
vec2 RaycastBox(vec3 ray_dir, vec3 minv, vec3 maxv)
{
	vec3 invDir = vec3(1.0) / ray_dir;

	vec3 t1 = minv * invDir;
	vec3 t2 = maxv * invDir;

	vec3 vtmin = min(t1, t2);
	vec3 vtmax = max(t1, t2);

	float tmin = max(vtmin[0], max(vtmin[1], vtmin[2]));
	float tmax = min(vtmax[0], min(vtmax[1], vtmax[2]));

	tmin = max(tmin, 0.0);

	return vec2(tmin, tmax);
}

float GetBoxPenetration(vec3 ray_dir, vec3 minv, vec3 maxv, float maxt)
{
	vec2 tvec = RaycastBox(ray_dir, minv, maxv);
	return max(min(tvec[1], maxt) - tvec[0], 0.0);
}

#endif

void main()
{
	//clip_pixel();

	mat4 dither = mat4(
	0.0/16.0-0.5, 8.0/16.0-0.5, 2.0/16.0-0.5, 10.0/16.0-0.5,
	12.0/16.0-0.5, 4.0/16.0-0.5, 14.0/16.0-0.5, 6.0/16.0-0.5,
	3.0/16.0-0.5, 11.0/16.0-0.5, 1.0/16.0-0.5, 9.0/16.0-0.5,
	15.0/16.0-0.5, 7.0/16.0-0.5, 13.0/16.0-0.5, 5.0/16.0-0.5);

	vec2 uv = gl_FragCoord.xy * ofs.xy;
	float storedDepth = texture(tex5, uv).x;

	float linear_depth = LinearDepth(storedDepth);
	vec3 viewPos = (corner + gl_FragCoord.xyz*cornerDelta)*linear_depth;

#ifdef FOG_DISTANCE
	float scale = length(viewPos);
#elif defined(FOG_SPHERE)
	float scale = GetSpherePenetration(normalize(world_dir), fog_pos_rel, fog_extent, length(viewPos));
#else
	float scale = GetBoxPenetration(normalize(world_dir), fog_box_min_rel, fog_box_max_rel, length(viewPos));
#endif

	scale *= fog_density;
	scale = dot(vec2(scale, scale*scale), fog_falloff);

	scale = clamp(1.0 - exp(-scale), 0.0, 1.0);

	fragColor = light_color;
	fragColor.w *= scale;

	ivec2 iv = ivec2(mod(gl_FragCoord.xy, vec2(4.0)));
	float dith = dither[iv.y][iv.x] * (1.0/255.0);

	fragColor.w += dith;
}
